
extends KinematicBody2D

var effectWallJump = preload("res://scenes/effects/walljump.scn")
var effectGroundJump = preload("res://scenes/effects/groundjump.scn")
var effectSnowStep = preload("res://scenes/effects/snowstep.scn")

const STATE_PLAYING = 1
const STATE_WIN = 2
const STATE_DEAD = 4
const STATE_WAIT = 8

const GRAVITY = 45.0
const JUMPPOWER = 27.5
const SPEED = 50.0
const DAMPING = Vector2(0.75, 0.9)
const WALLHANGTIME = 0.2
const MAX_INVULNERABLE_TIME = 0.5

var state = 1

var velocity = Vector2()
var collisionNormal = Vector2()

var health = 20
var invulnerable = 0
var dead = false

var jumping = false
var jumpsWall = 0
var jumpReleased = false
var jumpTime = 0
var lastWallJump = 0

var wallPress = 0
var wallTime = 0

var falldown = false
var onMovingPlatform = false
var onSnow = true

var leftFacing = false
var anim = "idle"


func take_damage(damage, origin=Vector2()):
	if invulnerable <= 0 and not dead:
		health = clamp(health - ceil(damage), 0, 20)
		if health <= 0:
			get_node("SamplePlayer").play("die")
			dead = true
			state = STATE_DEAD
		else:
			get_node("SamplePlayer").play("hurt")
		
		invulnerable = MAX_INVULNERABLE_TIME
		
		if origin.length() > 0:
			velocity = (get_global_pos() - origin).normalized() * GRAVITY * JUMPPOWER * 0.5
			velocity.y = -abs(velocity.y)
			wallTime = 0


func spawn():
	state = STATE_WAIT
	
	velocity = Vector2()
	collisionNormal = Vector2()
	
	health = clamp(floor((health + 20) / 2.0), 0, 20)
	
	get_node("Camera2D").set("smoothing/enable", false)


func win():
	state = STATE_WIN


func start():
	state = STATE_PLAYING
	get_node("Camera2D").set("smoothing/enable", true)


func restart():
	state = 1
	
	velocity = Vector2()
	collisionNormal = Vector2()
	
	health = 20
	invulnerable = 0
	dead = false
	
	jumping = false
	jumpsWall = 0
	jumpReleased = false
	jumpTime = 0
	lastWallJump = 0
	
	wallPress = 0
	wallTime = 0
	
	falldown = false
	onMovingPlatform = false
	onSnow = true
	
	leftFacing = false
	anim = "idle"


func snow_step_effect():
	get_node("SamplePlayer").play("walk")
	if onSnow:
		var e = effectSnowStep.instance()
		e.set_pos(get_pos() + Vector2(0, 14) + (velocity.normalized() * 16))
		get_parent().add_child(e)


func on_ground():
	if(collisionNormal.length() > 0):
		var floorHit = collisionNormal.dot(Vector2(0, -1)) >= 0.5
		return floorHit
	
	return false


func on_wall():
	if collisionNormal.length() > 0:
		var wallHit = collisionNormal.dot(Vector2(1, 0)) >= cos(deg2rad(15)) or collisionNormal.dot(Vector2(-1, 0)) >= cos(deg2rad(15))
		return wallHit
	
	return false


func _ready():
	set_fixed_process(true)
	set_process(true)


func _fixed_process(delta):
	
	var moveRight = Input.is_action_pressed("MOVE_RIGHT")
	var moveLeft = Input.is_action_pressed("MOVE_LEFT")
	var moveDown = Input.is_action_pressed("MOVE_DOWN")
	var moveJump = Input.is_action_pressed("MOVE_JUMP")
	
	var joyRightLeft = 0.05 * round(Input.get_joy_axis(0, 0) / 0.05)
	var joyUpDown = Input.get_joy_axis(0, 1)
	
	var isMovingRight = moveRight or joyRightLeft >= 0.2
	var isMovingLeft = moveLeft or joyRightLeft <= -0.2
	var isMovingDown = moveDown or joyUpDown >= 0.5
	
#	Movement
	velocity.y += GRAVITY
	
	if state == STATE_PLAYING:
		if invulnerable <= MAX_INVULNERABLE_TIME * 0.5:
			if isMovingRight:
				var f = 1
				if joyRightLeft >= 0.2:
					f = abs(joyRightLeft)
				velocity.x += SPEED * f
				leftFacing = false
			
			if isMovingLeft:
				var f = 1
				if joyRightLeft <= -0.2:
					f = abs(joyRightLeft)
				velocity.x -= SPEED * f
				leftFacing = true
		
#		Jumping
		if moveJump and not jumping and on_ground():
			get_node("SamplePlayer").play("jump")
			jumpTime = JUMPPOWER * delta
			jumping = true
			velocity.y = -GRAVITY * JUMPPOWER * jumpTime
			
			var e = effectGroundJump.instance()
			e.set_pos(get_pos() + Vector2(0, 14))
			e.set_on_snow(onSnow)
			get_parent().add_child(e)
			
		elif moveJump and jumpTime > 0 and not jumpReleased and not on_ground():
			jumpTime -= delta
			velocity.y = -GRAVITY * JUMPPOWER * jumpTime
			
		elif (not moveJump or jumpTime <= 0) and not on_ground():
			jumpReleased = true
			jumpTime = 0
		
		if wallTime > 0:
			wallTime -= delta
			velocity.x = wallPress
	
		#Wall Jump
		if moveJump and not jumping and on_wall() and not on_ground():
			get_node("SamplePlayer").play("jump")
			jumpReleased = false
			jumping = true
			
			if sign(wallPress) != lastWallJump:
				jumpsWall = 0
			
			var power = 1 / (1 + (jumpsWall * 2.0 / 3.0))
			jumpTime = (JUMPPOWER * delta + 0.1) * power
			
			lastWallJump = sign(velocity.x)
			
			velocity.x = sign(-velocity.x) * SPEED * 10
			velocity.y = -GRAVITY * JUMPPOWER * power
			
			#wall jump effect
			if power >= 1:
				var e = effectWallJump.instance()
				e.set_pos(get_pos() - Vector2(sign(velocity.x) * 8, 0))
				e.set_direction(-velocity.x)
				get_parent().add_child(e)
			
			jumpsWall += 1
		
		get_node("SamplePlayer").get_sample_library().sample_set_pitch_scale("jump", clamp(1.0 - (jumpsWall / 20.0), 0.5, 1.0))
	
#	Motion
	var onGroundBeforeMove = on_ground()
	
	if on_wall():
		if velocity.y > 0:
			velocity.y *= 0.65
		
		jumping = moveJump
		
		#Stick to wall
		wallPress = velocity.x
		if (isMovingLeft and wallPress < 0) or (isMovingRight and wallPress > 0):
			wallTime = WALLHANGTIME
		
		if wallTime > 0:
			if wallPress < 0:
				leftFacing = true
			else:
				leftFacing = false
		
	else:
		wallTime = 0
	
	if on_ground():
		jumpReleased = false
		jumpsWall = 0
		jumping = moveJump
		lastWallJump = 0
		wallTime = 0
		onSnow = true
		
		#lessen sliding on slopes, I don't know how to do this any better. Optimized for 45 degree slopes.
		velocity.x -= collisionNormal.x * GRAVITY * (1 / sin(collisionNormal.dot(Vector2(0, -1)))) * 0.9185
	
	#raycast below player
	var spaceState = get_world_2d().get_direct_space_state()
	var hit = spaceState.intersect_ray(get_pos() - Vector2(0, -16), get_pos() - Vector2(0, -17), [self])
	if not hit.empty():
		#fall through one way platforms
		if hit.collider.get_groups().find("platform") != -1:
			if isMovingDown and not falldown:
				set_pos(get_pos() - Vector2(0, -16))
			falldown = moveDown
			onSnow = false
		
		#move with moving platforms
		if hit.collider.get_groups().find("moving platform") != -1:
			move(hit.collider.get_parent().velocity)
			onMovingPlatform = true
			onSnow = false
		else:
			onMovingPlatform = false
	
	var motion = velocity * delta
	motion = move(motion)
	
#	Collision
	if is_colliding():
		collisionNormal = get_collision_normal()
		motion = collisionNormal.slide(motion)
		velocity = collisionNormal.slide(velocity)
		
		if not on_ground() and not on_wall():
			jumpTime = 0
		
		move(motion)
		
		set_pos(Vector2(get_pos().x, round(get_pos().y)))
		
	else:
		collisionNormal = Vector2()
	
	# Fix the collision vector from flipping out when pressing against a wall while still on the ground.
	if onGroundBeforeMove != on_ground() and on_wall() and not jumping:
		collisionNormal = Vector2(0,-1)
	
	velocity *= DAMPING
	
	invulnerable = max(invulnerable - delta, 0)


func _process(delta):
	var newAnim = anim
	var animPlayer = get_node("Anim")
	
	if leftFacing and velocity.x < 0:
		get_node("Sprite").set_scale(Vector2(-1, 1))
	elif not leftFacing and velocity.x > 0:
		get_node("Sprite").set_scale(Vector2(1, 1))
	
	animPlayer.set_speed(1)
	if state != STATE_DEAD:
		if abs(velocity.x) <= 1:
			newAnim = "idle"
			
			if state == STATE_PLAYING:
				if Input.is_action_pressed("MOVE_DOWN") or Input.get_joy_axis(0, 1) >= 0.5:
					newAnim = "crouch"
			
		else:
			if on_ground():
				newAnim = "walk"
				animPlayer.set_speed(abs(velocity.x) * 2.0 / 180.0)
		
		if wallTime > 0 and on_wall():
			newAnim = "wallslide"
		
		if not on_ground() and not on_wall():
			if velocity.y < 0:
				newAnim = "jump"
				animPlayer.set_speed(2.5)
			else:
				newAnim = "fall"
		
	else:
		newAnim = "dead"
	
	if newAnim != anim:
		anim = newAnim
		animPlayer.play(anim)
	
	get_node("Sprite").get_material().set_shader_param("StunAmount", invulnerable / MAX_INVULNERABLE_TIME)


